# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
UP_VERB = "up"
DOWN_VERB = "down"
GET_VERB = "get"
METADATA_VERB = "metadata"
DATABAG_FILE = "databag.yaml"
OUTPUT_DATABAG_FILE = f"output_{DATABAG_FILE}"

STACK_DIR = "stack"
USER_DIR = "user"

# if we download files, put them inside this directory so we know to skip
SKIP_DIR = "download"

PATTERN_LOCAL_CLOUDFORMATION_FILE = ".cloudformation.yaml"
PATTERN_REMOTE_CLOUDFORMATION_FILE = ".remote_cloudformation.yaml"
PATTERN_BASH = ".sh"
PATTERN_KUBECTL_FILE = ".kubectl.yaml"
PATTERN_SOLARWINDS_PAPERTRAIL_FILE = "solarwinds_papertrail.yaml"
PATTERN_KUSTOMIZATION_FILE = "kustomization.yaml"
PATTERN_RINGMASTER_PYTHON_FILE = ".ringmaster.py"
PATTERN_SNOWFLAKE_SQL = ".snowflake.sql"
PATTERN_SNOWFLAKE_QUERY = ".snowflake_query.sql"
PATTERN_HELM_DEPLOY = "helm_deploy.yaml"
PATTERN_AWS_IAM_POLICY = ".iam_policy.json"
PATTERN_AWS_IAM_ROLE = ".iam_role.json"
PATTERN_SECRETS_MANAGER = "secretsmanager.yaml"
PATTERN_EKSCTL_CONFIG = ".eksctl.yaml"
PATTERN_SECRET_KUBECTL = ".secret_kubectl.yaml"
PATTERN_CLOUDFLARE = ".cloudflare.yaml"

MSG_UP_TO_DATE = "[√] up to date"
SUBSTITUTE_VARIABLE_REGEX = r"(\$\{[^}]+})"
SINGLE_QUOTED_STRING_REGEX = r"('.*?(?<!\\)')"

KEY_INTERMEDIATE_DATABAG="intermediate_databag_file"

AWS_TEMPLATE_DIR = "res/aws"
AWS_USER_TEMPLATE_DIR = "~/.ringmaster/res/aws"

COMMENT_SQL = "--"
SNOWFLAKE_CLEANUP_FILENAME = "down.snowflake.sql"


CFN_BASE64 = "base64"
ENV_DIR = ".env"
METADATA_FILE = "metadata.yaml"
METADATA_FILES_KEY = "files"
METADATA_NAME_KEY = "name"
METADATA_HASH_KEY = "sha1sum"
SOURCE_KEY = "source"

DEFAULT_DATABAG = {
    "msg_up_to_date": MSG_UP_TO_DATE,
    "up_verb": UP_VERB,
    "down_verb": DOWN_VERB,
    "name": os.path.basename(os.getcwd()),
}

PROCESSED_DIR = ".processed"
DATABAG_ENV_KEY = "env_name"
CONNECTIONS_YAML = "connections.yaml"
PROFILE = "profile"


# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from urllib.parse import urlparse, urlunparse
import importlib
import requests
import os
import glob
import yaml
import json
import sys
import tempfile
from loguru import logger
from ringmaster import constants as constants
import ringmaster.k8s as k8s
import ringmaster.aws as aws
import ringmaster.snowflake as snowflake
from pathlib import Path
import ringmaster.version as version
import ringmaster.util as util
import ringmaster.cloudflare as cloudflare

debug = False

metadata = {
    "ringmaster_version": version.__version__,
    "generated_at": datetime.now().isoformat(),
    "description": "edit me",
    constants.METADATA_FILES_KEY: {},
}

# munged directory containing databag user supplied on the commandline
env_dir = None


def init_databag():
    """per-run program specific data"""

    # users write values as JSON to this file and they are added to the
    # databag incrementally
    _, intermediate_databag_file = tempfile.mkstemp(suffix="json", prefix="ringmaster")

    return {
        constants.KEY_INTERMEDIATE_DATABAG: intermediate_databag_file,
        "debug": "debug" if debug else "",
    }


def load_databag(databag_file):
    # load values from user
    data = {}
    if os.path.exists(databag_file):
        with open(databag_file) as f:
            data.update(yaml.safe_load(f))
    else:
        logger.warning(f"missing databag file: {databag_file}")

    logger.debug(f"loaded databag contents: {data}")
    return data


def load_intermediate_databag(data):
    intermediate_databag_file = data[constants.KEY_INTERMEDIATE_DATABAG]
    if os.path.getsize(intermediate_databag_file):
        # grab stashed databag
        logger.debug(f"loading databag left by last stage from {intermediate_databag_file}")
        with open(intermediate_databag_file, "r") as json_file:
            extra_data = json.load(json_file)

        logger.debug(f"loaded {len(extra_data)} items: {extra_data}")
        data.update(extra_data)

        # empty the bag for next iteration
        open(intermediate_databag_file, 'w').close()


def do_bash_script(working_dir, filename, verb, data):
    # bash scripts
    logger.info(f"bash script: {filename}")
    logger.info(f"write JSON to file at ${constants.KEY_INTERMEDIATE_DATABAG} to add to databag")
    util.run_cmd(f"bash {filename} {verb}", data)

    load_intermediate_databag(data)


# see
# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
def do_ringmaster_python(working_dir, filename, verb, data):
    logger.info(f"ringmaster python: {filename}")
    # /foo/bar/baz.py -> baz
    module_name, _ = os.path.splitext(os.path.basename(filename))
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    # load data and run plugin
    # logger.enable("storage_security_groups.ringmaster")
    # logger.enable("storage_security_groups")
    # logger.enable("__main__")
    # logger.enable(module_name)
    # module.loggey = logger
    module.databag = data
    module.main(verb)


handlers = {
    constants.PATTERN_BASH: do_bash_script,
    constants.PATTERN_LOCAL_CLOUDFORMATION_FILE: aws.do_local_cloudformation,
    constants.PATTERN_REMOTE_CLOUDFORMATION_FILE: aws.do_remote_cloudformation,
    constants.PATTERN_KUBECTL_FILE: k8s.do_kubectl,
    constants.PATTERN_KUSTOMIZATION_FILE: k8s.do_kustomizer,
    constants.PATTERN_RINGMASTER_PYTHON_FILE: do_ringmaster_python,
    constants.PATTERN_SNOWFLAKE_SQL: snowflake.do_snowflake_sql,
    constants.PATTERN_SNOWFLAKE_QUERY: snowflake.do_snowflake_query,
    constants.PATTERN_HELM_DEPLOY: k8s.do_helm,
    constants.PATTERN_AWS_IAM_POLICY: aws.do_iam_policy,
    constants.PATTERN_AWS_IAM_ROLE: aws.do_iam_role,
    constants.PATTERN_SECRETS_MANAGER: aws.do_secrets_manager,
    constants.PATTERN_EKSCTL_CONFIG: aws.do_eksctl,
    constants.PATTERN_SECRET_KUBECTL: k8s.do_secret_kubectl,
    constants.PATTERN_CLOUDFLARE: cloudflare.do_cloudflare,
}


def get_handler_for_file(filename):
    handler = None
    for pattern in  handlers.keys():
        if filename.endswith(pattern):
            handler = handlers[pattern]
            break
    return handler



def do_file(working_dir, filename, verb, data):
    handler = get_handler_for_file(filename)
    if handler and verb == constants.METADATA_VERB:
        metadata[constants.METADATA_FILES_KEY][os.path.basename(filename)] = {
            constants.METADATA_HASH_KEY: util.hash_file(filename)
        }
    elif handler:
        handler(working_dir, filename, verb, data)
    else:
        logger.debug(f"no handler for {filename} - skipped")


def do_stage(working_dir, data, stage, verb):
    """process the children of an 'inner' dir in order, eg:
    stacks/
        0010 <--- this level
            somefile.yaml
    """
    for root, dirs, files in os.walk(stage, topdown=True, followlinks=True):
        # modify dirs in-place to exclude dirs to skip and all their children
        # https://stackoverflow.com/a/19859907/3441106
        dirs[:] = list(filter(lambda x: not x == constants.SKIP_DIR, dirs))

        # and then sort the files...
        files.sort()
        for file in files:
            filename = os.path.join(root, file)
            do_file(working_dir, filename, verb, data)

        if verb != constants.METADATA_VERB:
            save_output_databag(data)


def system_info():
    aws_version = util.run_cmd(["aws", "--version"]).strip()
    eksctl_version = util.run_cmd(["eksctl", "version"]).strip()
    kubectl_version = util.run_cmd(["kubectl", "version", "--short", "--client"]).strip()
    helm_version = util.run_cmd(["helm", "version", "--short"])

    message = util.process_res_template(
        "system_info.txt",
        aws_version=aws_version,
        boto3_version=aws.boto3_version(),
        cloudflare_version=cloudflare.cloudflare_version(),
        eksctl_version=eksctl_version,
        helm_version=helm_version,
        kubectl_version=kubectl_version,
        snowflake_connector_version=snowflake.snowflake_connector_version()
    )

    logger.info(message)


def setup_connections():

    connections = get_env_connections()
    aws.setup_connection(connections.get("aws"))
    snowflake.setup_connection(connections.get("snowflake"))
    k8s.setup_connection(connections.get("k8s"))

    # no connection info for cloudflare - you can only manage a DNS zone once
    # so this will either work or it wont...

def run(project_dir, filename, merge, env_name, verb):
    if os.path.exists(filename):
        data = get_env_databag(os.getcwd(),merge, env_name)
        setup_connections()

        do_file(project_dir, filename, verb, data)
        save_output_databag(data)
    else:
        logger.error(f"file not found: {filename}")


def delete_output_databag():
    output_databag_file = get_output_databag_filename()
    if os.path.exists(output_databag_file):
        logger.debug(f"deleting databag: {output_databag_file}")
        os.unlink(output_databag_file)


def save_output_databag(data):
    output_databag_file = get_output_databag_filename()
    logger.info(f"saving output databag:{output_databag_file}")
    util.save_yaml_file(
        output_databag_file,
        data,
        "# generated by ringmaster, do not edit!\n"
    )


def run_dir(working_dir, subdir, merge, env_name, start, verb):
    """process an 'outer' dir and all its children in order, eg:
    stacks/ <--- this level
        0010
            somefile.yaml
    """
    if os.path.exists(subdir):
        logger.debug(f"found: {subdir}")
        started = False

        data = get_env_databag(os.getcwd(), merge, env_name)
        setup_connections()

        # for some reason the default order is reversed when using ranges so we
        # must always sort. If we are bringing down a stack, reverse the order
        # to process steps last->first - dont rely on strange behaviour
        stages = sorted(glob.glob(f"./{subdir}/[0-9][0-9][0-9][0-9]*"), reverse=(verb == constants.DOWN_VERB))
        first_dir = os.path.basename(stages[0])
        last_dir = os.path.basename(stages[-1])
        if not start:
            start = first_dir if constants.UP_VERB else last_dir
            logger.debug(f"setting start dir:{start}")

        for stage in stages:
            logger.debug(stage)
            if not started:
                number = os.path.basename(stage)
                if number == start:
                    started = True

            if started:
                logger.debug(f"stage: {stage}")
                do_stage(working_dir, data, stage, verb)

        if not started:
            logger.error(f"start dir - not found: {start}")

        # cleanup
        logger.debug("delete intermediate databag")
        os.unlink(data[constants.KEY_INTERMEDIATE_DATABAG])

        if verb == constants.DOWN_VERB:
            delete_output_databag()

    else:
        logger.error(f"missing directory: {subdir}")


def check_ok_to_update(local_metadata_file, remote_url):
    safe = True
    working_dir = os.path.dirname(local_metadata_file)
    if os.path.exists(local_metadata_file):
        logger.debug(f"loading metadata:{local_metadata_file}")
        with open(local_metadata_file) as f:
            local_metadata = yaml.safe_load(f)
            logger.debug(f"metadata loaded:{local_metadata}")
        local_url = local_metadata.get(constants.SOURCE_KEY)
        if local_url == remote_url:
            for filename, file_metadata in local_metadata.get(constants.METADATA_FILES_KEY, {}).items():
                local_filename = os.path.join(working_dir, filename)
                if os.path.exists(local_filename) and \
                        util.hash_file(local_filename) != file_metadata[constants.METADATA_HASH_KEY]:
                    logger.error(f"file:{local_filename} MODIFIED - aborting. Delete and retry to overwrite")
                    safe = False
        else:
            logger.error(f"file:{local_metadata_file} source:{local_url} does not match requested:{remote_url} - aborting. Try saving to another directory")
            safe = False
    return safe


def download_metadata_yaml(url):
    # append METADATA_FILE to the end of path if missing...
    metadata_url = util.change_url_filename(url, constants.METADATA_FILE)
    return yaml.safe_load(requests.get(metadata_url).text)


def download_files_from_metadata(directory, new_metadata, base_url):
    """ download each file listed in new_metadata['files'] to `directory` """
    for filename, file_metadata in new_metadata.get(constants.METADATA_FILES_KEY, {}).items():
        local_file = os.path.join(directory, filename)
        remote_url = util.change_url_filename(base_url, filename)

        logger.info(f"downloading {remote_url} ==> {local_file}")
        downloaded_file = requests.get(remote_url)
        with open(local_file, "wb") as f:
            f.write(downloaded_file.content)

        if not util.hash_file(local_file) == new_metadata[constants.METADATA_FILES_KEY][filename][constants.METADATA_HASH_KEY]:
            raise RuntimeError(f"local hash != remote hash file:{filename}")


def get(directory, url):
    """download <url> to <dir> - somewhat inspired by go get"""
    if not os.path.exists(directory):
        Path(directory).mkdir(parents=True, exist_ok=True)

    local_metadata_file = os.path.join(directory, constants.METADATA_FILE)
    if check_ok_to_update(local_metadata_file, url):
        new_metadata = download_metadata_yaml(url)
        download_files_from_metadata(directory, new_metadata, url)

        # save the metadata file and add the source we downloaded from
        new_metadata[constants.SOURCE_KEY] = url
        util.save_yaml_file(local_metadata_file, new_metadata)


def include_extra_file(directory, extra_file):
    extra_file_path = os.path.join(directory, extra_file)
    if os.path.exists(extra_file_path):
        metadata[constants.METADATA_FILES_KEY][extra_file] = {
            constants.METADATA_HASH_KEY: util.hash_file(extra_file_path)
        }
    else:
        raise RuntimeError(f"Requested --include {extra_file_path} not found")


def write_metadata(directory, includes):
    extra_files = includes.split(",") if includes else []
    metadata_file = os.path.join(directory, constants.METADATA_FILE)
    name = os.path.basename(os.path.abspath(directory))
    metadata[constants.METADATA_NAME_KEY] = name
    logger.info(f"Collecting metadata for {directory} to {metadata_file} (--includes:{extra_files})")
    if os.path.isdir(directory):
        do_stage({}, directory, constants.METADATA_VERB)

        # now add any extra includes
        for extra_file in extra_files:
            include_extra_file(directory, extra_file)
    else:
        raise RuntimeError(f"No such directory: {directory}")

    util.save_yaml_file(metadata_file, metadata)


def get_env_dir(working_dir, env_name):
    env_dir_section = env_name if env_name else ""
    dirs = list(
        filter(
            lambda x: x is not None,
            [working_dir, constants.ENV_DIR, env_dir_section]
        )
    )
    return os.path.join(*dirs)


def get_env_connections():
    """read `connections.yaml` for this environment"""
    if not env_dir:
        raise RuntimeError("env_dir not set yet, load databags first")

    connections_yaml_filename = os.path.join(env_dir, constants.CONNECTIONS_YAML)
    return util.read_yaml_file(connections_yaml_filename)


def get_env_databag(working_dir, merge, env_name):
    """read the databag for this `env_name` with optional merging of any
    parent databags"""
    global env_dir

    # deepest directory to load databag from, with optional merging
    # examples: .env, .env/prod, .env/prod/australia
    env_dir = get_env_dir(working_dir, env_name)

    sequential_load_dirs = []
    look_at_dir = env_dir
    max_parents = 4
    depth = 0
    if merge:
        # add parent directory until we get back to .env
        while look_at_dir != working_dir:
            if depth > max_parents:
                raise RuntimeError(f".env dir nested too deep! (limit: {max_parents+1}")
            sequential_load_dirs.append(look_at_dir)
            look_at_dir = os.path.dirname(look_at_dir)
            depth += 1
        sequential_load_dirs.reverse()
    else:
        sequential_load_dirs.append(env_dir)

    logger.debug(f"sequential load dirs: {sequential_load_dirs}")

    # shallow
    data = constants.DEFAULT_DATABAG.copy()
    for look_at_dir in sequential_load_dirs:
        databag_file = os.path.join(look_at_dir, constants.DATABAG_FILE)
        output_databag_file = os.path.join(look_at_dir, constants.OUTPUT_DATABAG_FILE)
        target_databag_file = output_databag_file \
            if os.path.exists(output_databag_file) else databag_file
        data.update(load_databag(target_databag_file))

    data.update(init_databag())
    # `env` will clash with scoped environment variables
    data[constants.DATABAG_ENV_KEY] = env_name
    return data


def get_output_databag_filename():
    if not env_dir:
        raise RuntimeError("databag_load_dir not set, databag not loaded yet")
    return os.path.join(env_dir, constants.OUTPUT_DATABAG_FILE)



# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import snowflake.connector
import os
from loguru import logger
import yaml
import ringmaster.util as util
import ringmaster.constants as constants

SNOWFLAKE_CONFIG_FILE = "~/.ringmaster/snowflake.yaml"
connection = None

profile_data = None


def setup_connection(connection_settings):
    global profile_data

    profile = util.get_connection_profile(connection_settings, "snowflake")
    snowflake_config_file = os.path.expanduser(SNOWFLAKE_CONFIG_FILE)
    if os.path.exists(snowflake_config_file):
        config = util.read_yaml_file(snowflake_config_file)
        profile_data = config.get(profile)
    else:
        raise RuntimeError(f"snowflake settings not found at: {snowflake_config_file}")


def get_cursor(data):
    ctx = snowflake.connector.connect(**profile_data["credentials"])
    cs = ctx.cursor(snowflake.connector.DictCursor)

    # grab some convenince variables from snowflake settings
    data["snowflake_account"] = profile_data["credentials"]["account"]
    data["snowflake_region"] = profile_data["region"]

    test_connection(cs)
    return cs


def test_connection(cs):
    cs.execute("SELECT current_version()")
    one_row = cs.fetchone()


def process_file_and_connect(working_dir, filename, verb, data):
    # process substitutions
    processed_file = util.substitute_placeholders_from_file_to_file(
        working_dir,
        filename,
        constants.COMMENT_SQL,
        verb,
        data
    )
    logger.debug(f"snowflake processed file: {processed_file}")

    # connect to snowflake, bail if it fails
    cs = get_cursor(data)
    return cs, processed_file


def do_snowflake_sql(working_dir, filename, verb, data):
    """Run an SQL script with substition variables"""
    script_name = os.path.basename(filename)
    if (verb == constants.UP_VERB and script_name != constants.SNOWFLAKE_CLEANUP_FILENAME) or \
            (verb == constants.DOWN_VERB and script_name == constants.SNOWFLAKE_CLEANUP_FILENAME):

        logger.info(f"snowflake sql: {filename}")
        cs, processed_file = process_file_and_connect(working_dir, filename, verb, data)

        # build up stmt line-by-line, when we find `;` execute stmt and
        # empty the variable for the next iteration
        stmt = ""
        with open(processed_file, "r") as file:
            for line in file:
                if not line.startswith(constants.COMMENT_SQL):
                    stmt += line.rstrip()
                    if stmt.endswith(";"):
                        logger.debug(f"sql: {stmt}")
                        cs.execute(stmt)
                        stmt = ""
    else:
        logger.info(f"skippking snowflake: {filename}")


def do_snowflake_query(working_dir, filename, verb, data):
    """Query snowflake for a single row of values, add each column to the databag"""
    if verb == constants.UP_VERB:
        logger.info(f"snowflake query: {filename}")
        cs, processed_file = process_file_and_connect(
            working_dir,
            filename,
            verb,
            data
        )
        extra_data = {}

        # load the entire processed file and run EACH STATEMENT
        with open(processed_file, 'r') as file:
            sql = file.read()

        for stmt in sql.split(";"):
            logger.debug(f"snowflake sql query: {stmt}")
            cs.execute(stmt)

        result = cs.fetchone()
        logger.info(f"sql result: {result}")

        # result is a dict so just lowercase each key and add to databag
        for k, v in result.items():
            extra_data[k.lower()] = v

        logger.debug(f"query result - items: {len(extra_data)} values:{extra_data}")
        data.update(extra_data)


def snowflake_connector_version():
    return snowflake.connector.__version__



# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import CloudFlare
import ringmaster.constants as constants
import ringmaster.util as util
import ringmaster.aws as aws
from loguru import logger
import yaml
import re
import tempfile
import os
import json


def get_cf():
    """
    Get an authenticated cloudflare API instance. Authentication resolved in
    order from:
    1. `.cloudflare.cfg`
    2. Environment variables

    @see
    https://github.com/cloudflare/python-cloudflare#providing-cloudflare-username-and-api-key
    """
    return CloudFlare.CloudFlare()


def list_contains_dict_value(data, key, target):
    found = False
    logger.debug(f"filtering list for dict value: {key}=={target}")
    for hostname in data.get(key, []):
        if re.search(f"^{re.escape(target)}$", hostname):
            found = True
            break

    return found


def create_csr(hostname):
    _, csr_temp = tempfile.mkstemp()
    _, private_key_temp = tempfile.mkstemp()
    cmd = f"""openssl \
        req -new -newkey rsa:2048 -nodes \
        -out {csr_temp} \
        -keyout {private_key_temp} \
        -subj "/C=US/ST=California/L=/O=/CN={hostname}"
    """
    util.run_cmd(cmd)

    # read both files into memory and return them, deleteing the tempfiles
    with open(csr_temp) as f:
        csr_data = f.read()

    with open(private_key_temp) as f:
        private_key_data = f.read()

    os.remove(csr_temp)
    os.remove(private_key_temp)

    return csr_data, private_key_data


def ensure_origin_ca_cert(cf, verb, origin_certs, hostname, cb):
    """create origin CA cert for `domain` if needed"""

    def hostname_filter(data):
        return list_contains_dict_value(data, "hostnames", hostname)
    pass

    origin_ca_cert = list(filter(hostname_filter, origin_certs))

    # We cannot update/create missing secrets for origins that have
    # already been provisioned as the certificate private key is only
    # available while the CSR is being processed. If secret needs to
    # autohealing capability would have to delete the existing origin
    # CA and sign a new cert.
    # TLDR - To re-create secret delete (revoke) the Origin CA cert and
    # the secret in secrets manager, then re-run
    certificate = None
    private_key_data = None
    if verb == constants.DOWN_VERB and origin_ca_cert:
        cert_id = origin_ca_cert[0]["id"]
        logger.info(f"Deleting Origin CA cert: {hostname}/{cert_id}")
        cf.certificates.delete(cert_id)
    elif verb == constants.UP_VERB and not origin_ca_cert:
        logger.debug(f"Creating Origin CA cert: {hostname}")
        csr_data, private_key_data = create_csr(hostname)
        res = cf.certificates.post(data={
            "hostnames": [hostname, f"*.{hostname}"],
            "csr": csr_data,
            "request_type": "origin-rsa",
        })
        certificate = res["certificate"]
        logger.info(res.get("certificate"))
        logger.info(res.get("csr"))
    else:
        logger.info(constants.MSG_UP_TO_DATE)

    logger.debug(f"processing callback for {hostname}")
    cb(verb, hostname, certificate, private_key_data)


# walkthru: https://blog.cloudflare.com/cloudflare-ca-encryption-origin/
def origin_ca_certs(cf, zone_id, verb, zone_data, cb):
    # create each requested cert (if stack is going down just delete
    # everything to avoid creating orphans
    certs_to_ensure = zone_data.get("origin_ca_certs", [])
    if certs_to_ensure:
        # get the list of all origin ca certs for this domain
        origin_certs = cf.certificates.get(params={"zone_id": zone_id})

        for domain in certs_to_ensure:
            logger.debug(f"ensure wildcard origin CA cert: {domain}")
            ensure_origin_ca_cert(cf, verb, origin_certs, domain, cb)


def get_zone_id(cf, zone_name):
    """get the zone_id for a zone name or `False` if not exist"""
    params = {"name": zone_name, "per_page": 1}
    zones = cf.zones.get(params=params)
    try:
        zone_id = zones[0]['id']
    except KeyError:
        raise RuntimeError(f"zone {zone_name} not visible in cloudflare! - check setup/permissions")

    logger.debug(f"cloudflare resolved zone {zone_name} -> {zone_id}")
    return zone_id


def zone_settings(cf, zone_id, verb, yaml_data):
    """zone-wide cloudflare settings (strict mode ssl)"""

    # list of dict -> dict
    if verb == constants.UP_VERB:
        settings_list_of_dict = cf.zones.settings.get(zone_id)
        settings_dict = {item['id']: item["value"] for item in settings_list_of_dict}
        logger.debug(f"settings for zone_id {zone_id}: {settings_dict}")
        for key, value in yaml_data.get("settings", {}).items():
            if settings_dict.get(key) == value:
                logger.debug(f"up-to-date: {key}=>{value}")
            else:
                logger.debug(f"setting: {key}=>{value}")
                getattr(cf.zones.settings, key).patch(zone_id, data={"value": value})
    else:
        logger.info("skipping cloudformation zone settings")


def edge_certs(cf, zone_id, verb, yaml_data):
    """create or delete the named edge certificates. Edge certs are 100%
    managed by cloudflare so the orgin never knows about or sees them"""
    certs = cf.zones.ssl.certificate_packs.get(zone_id)

    for hostname in yaml_data.get("edge_certs", []):
        logger.debug(f"edge cert: {hostname}")

        def hostname_filter(data):
            return list_contains_dict_value(data, "hosts", hostname)
        pass

        issued = list(filter(hostname_filter, certs))
        if verb == constants.UP_VERB and not issued:
            logger.info(f"[cloudflare] creating edge certificate: {hostname}")
            cf.zones.ssl.certificate_packs.order.post(zone_id, data={
                "type": "advanced",
                "hosts": [yaml_data["zone_name"], hostname],
                "validation_method": "txt",
                "validity_days": 365,
                "certificate_authority": "digicert",
            })
        elif verb == constants.DOWN_VERB and issued:
            cert_id = issued[0]["id"]
            logger.info(f"[cloudflare] deleting edge certificate: {hostname}/{cert_id}")
            cf.zones.ssl.certificate_packs.delete(zone_id, cert_id)
        else:
            logger.info(constants.MSG_UP_TO_DATE)


def do_cloudflare(working_dir, filename, verb, data=None):
    logger.info(f"cloudflare: {filename}")

    try:
        processed_file = util.substitute_placeholders_from_file_to_file(
            working_dir,
            filename,
            "#",
            verb,
            data
        )
        yaml_data = util.read_yaml_file(processed_file)
        logger.debug(f"cloudflare file processed OK")

        prefix = yaml_data.get("aws").get("secrets_manager_prefix") or ""
        logger.debug(f"aws secretsmanager prefix: {prefix}")

        # callback to run when a certificate has been created
        # for now, always creates an AWS secret
        def cb(_verb, hostname, certificate_data, private_key_data):
            secret_string = json.dumps({
                    "tls.crt": certificate_data,
                    "tls.key": private_key_data,
                }
            )
            secret = {
                "name": f"{prefix}tls-{hostname.replace('.', '-')}",
                "value": secret_string,
            }

            if (_verb == constants.UP_VERB and certificate_data and private_key_data) \
                    or _verb == constants.DOWN_VERB:
                aws.ensure_secret(data, _verb, secret)
            else:
                logger.info(f"[cloudflare-aws-callback] no update required for: {hostname}")

        # lookup zone id now as needed everywhere
        cf = get_cf()
        zone_id = get_zone_id(cf, yaml_data["zone_name"])

        origin_ca_certs(cf, zone_id, verb, yaml_data, cb)
        edge_certs(cf, zone_id, verb, yaml_data)
        zone_settings(cf, zone_id, verb, yaml_data)
    # except RuntimeError as e:
    #     if verb == constants.DOWN_VERB:
    #         logger.warning(f"kubectl error - moving on: {e}")
    #     else:
    #         raise e
    except KeyError as e:
        if verb == constants.DOWN_VERB:
            logger.warning(f"missing key - moving on: {e}")
        else:
            raise e


def cloudflare_version():
    return CloudFlare.__version__


# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from . import constants
from loguru import logger
import subprocess
import requests
import snakecase
import json
from contextlib import ExitStack
from halo import Halo
import hashlib
import base64
from jinja2 import Environment, Template, StrictUndefined, Undefined
from jinja2.exceptions import UndefinedError
import yaml
import pathlib


def walk(data, parent_name=None):
    seq_iter = data if isinstance(data, dict) else range(len(data))
    for i in seq_iter:
        if parent_name:
            me = f"{parent_name}_{i}"
        else:
            me = i

        if isinstance(data[i], dict):
            for k, v in walk(data[i], parent_name=me):
                yield k, v
        elif isinstance(data[i], list) or isinstance(data[i], tuple):
            for k, v in walk(data[i], parent_name=me):
                yield k, v
        else:
            yield me, data[i]


def convert_dict_values_to_string(data):
    # all values passed to os.environ must be strings, avoid unwanted yaml help
    for key in data:
        data[key] = str(data[key])


def merge_env(data):
    env = os.environ.copy()
    env.update(data)

    convert_dict_values_to_string(env)
    return env


def run_cmd_json(cmd, data=None):
    """Run a command and parse JSON from its output"""
    string = run_cmd(cmd, data=data)
    logger.debug(f"string to parse: {string}")
    return json.loads(string)


def run_cmd(cmd, data=None):
    if not data:
        data = {}
    output = ""
    env = merge_env(data)
    logger.trace(f"merged environment: {env}")
    logger.debug(f"running command: {cmd}")
    debug = data.get("debug", False)
    with ExitStack() as stack:
        if not debug:
            stack.enter_context(Halo(text=f"Running {cmd}", spinner='dots'))

        with subprocess.Popen(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT,
                              shell=isinstance(cmd, str),
                              env=env) as proc:
            while True:
                line = proc.stdout.readline()
                while line:
                    line_decoded = line.decode("UTF-8")
                    if debug:
                        logger.log("OUTPUT", line_decoded.strip())
                    output += line_decoded

                    line = proc.stdout.readline()
                if proc.poll() is not None:
                    break
        rc = proc.poll()
        logger.debug(f"result: {rc}")
        if rc != 0:
            logger.error(output)
            raise RuntimeError(f"Command failed with non-zero exit status:{rc} - {cmd}")
    return output


def flatten_nested_dict(data):
    flattened = {}
    for k, v in walk(data):
        flattened[k.lower()] = v

    return flattened


def base64encode(string):
    string_bytes = string.encode('ascii')
    base64_bytes = base64.b64encode(string_bytes)
    return base64_bytes.decode('ascii')


def get_res_filename(filename):
    return os.path.join(
        os.path.dirname(
            os.path.realpath(__file__)
        ),
        "res",
        filename
    )


def get_res_file_content(filename):
    with open(get_res_filename(filename), "r") as f:
        return f.read()


def process_res_template(filename, **data):
    """process a template from the /res directory and return it as a string"""
    template = Template(get_res_file_content(filename))
    return template.render(**data)


def substitute_placeholders_from_memory_to_memory(raw, verb, data):
    """replace all variables placeholders list of lines and return the result"""

    # allow missing variables in templates if we are going down
    undefined = StrictUndefined if verb == constants.UP_VERB else Undefined
    jinja_env = Environment(undefined=undefined, keep_trailing_newline=True)
    # compatible name with Ansible filters
    jinja_env.filters['b64encode'] = base64encode
    template = jinja_env.from_string(raw)
    # add `env` key with contents of environment
    data_with_env = {**data, "env": os.environ.copy()}

    try:
        buffer = template.render(data_with_env)
    except UndefinedError as e:
        if verb == constants.DOWN_VERB:
            logger.warning(f"returning original content due to: {e}")
            buffer = raw
        else:
            raise e
    return buffer


def substitute_placeholders_from_file_to_memory(filename, verb, data):
    """replace all variables placeholders in filename and return the result"""
    if os.path.exists(filename):
        with open(filename, "r") as in_file:
            buffer = substitute_placeholders_from_memory_to_memory(
                in_file.read(),
                verb,
                data
            )
    else:
        raise RuntimeError(f"No such file: {filename}")

    return buffer


def get_processed_filename(working_dir, input_filename, env_name):
    """Return the filename where we should save the processed version of
    `input_filename`"""
    if input_filename.startswith("/"):
        raise RuntimeError(f"input_filename must be a relative path, received: {input_filename}")

    dirs = filter(
        lambda x: x is not None,
        [working_dir, constants.PROCESSED_DIR, env_name, input_filename]
    )
    processed_filename = os.path.join(*dirs)

    if os.path.abspath(input_filename) == os.path.abspath(processed_filename):
        raise RuntimeError(f"filename and processed_filename refer to the same file: {input_filename}")

    return os.path.normpath(processed_filename)


def substitute_placeholders_from_file_to_file(working_dir, filename, comment_delim, verb, data):
    """replace all variables placeholders in filename, return path to substituted file"""
    processed_file = get_processed_filename(working_dir, filename, data.get(constants.DATABAG_ENV_KEY))
    logger.debug(f"substitute placeholders: {filename} => {processed_file}")
    abs_filename = os.path.normpath(os.path.join(working_dir, filename))
    if os.path.exists(abs_filename):
        # create parent directory for processed file if needed
        pathlib.Path(os.path.dirname(processed_file)).mkdir(parents=True, exist_ok=True)
        with open(processed_file, "w") as out_file:
            out_file.write(f"{comment_delim} This file was automatically generated from file: {filename}, do not edit!\n")
            with open(abs_filename, "r") as in_file:
                out_file.write(
                    substitute_placeholders_from_memory_to_memory(
                        in_file.read(),
                        verb,
                        data,
                    )
                )
    else:
        raise RuntimeError(f"No such file: {abs_filename}")
    return processed_file


def download(url, filename):
    downloaded = requests.get(url, allow_redirects=True)
    open(filename, 'wb').write(downloaded.content)


# convert `number` to `_number` to match databag
def string_to_snakecase(string):
    # there are two conversion patterns in use which convert as follows:
    #   1. lowercase-hypen-separated -> lowercase_hyphen_separated
    #   2. mixed-CasePascalCaseAndHyphenSeparated -> mixed_case_pascal_case_and_hyphen_separated
    # AWS public cloudformation scripts use option 2 ;-)
    if string.lower() == string:
        # lower case
        string = string.replace("-", "_")
    else:
        # mixed case
        string = string.replace("-", "")

    key = snakecase.convert(string)

    logger.debug(f"converted input:{string} key:{key}")
    return key


def hash_file(filename):
    return hashlib.sha1(pathlib.Path(filename).read_bytes()).hexdigest()


def change_url_filename(url, filename):
    """change the filename in the last part of url if needed, eg:
        change_url_filename("https://blah", "something.txt") # https://blah/something.txt
    """
    return f"{url}/{filename}"


def read_yaml_file(filename):
    """read a yaml file and return it"""
    with open(filename) as f:
        yaml_data = yaml.safe_load(f)

    if yaml_data is None:
        raise RuntimeError(f"No YAML data in file: {filename}")

    return yaml_data


def save_yaml_file(filename, data, comment=None):
    """save yaml data to file, creating any directories as needed"""
    dirname = os.path.dirname(filename)
    pathlib.Path(dirname).mkdir(parents=True, exist_ok=True)
    with open(filename, "w") as f:
        if comment:
            f.write(comment)
        yaml.dump(data, f)


def get_connection_profile(connection, caller_name):
    """get the kubectl context to run with context set or bomb out"""
    if not connection.get(constants.PROFILE):
        raise RuntimeError(f"No profile set for {caller_name}")

    return connection[constants.PROFILE]


# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import subprocess
import os
import json
from loguru import logger
import boto3
import snakecase
import pathlib
import ringmaster.util as util
from ringmaster import constants as constants
from cfn_tools import load_yaml
import botocore.exceptions
from halo import Halo
from ringmaster.util import flatten_nested_dict

# AWS/boto3 API error messages to look for. Use partial regex to protect
# against upstream changes as much as we can
ERROR_UP_TO_DATE = r"No updates are to be performed"
ERROR_AWS = r"encountered a terminal failure state"
ERROR_MISSING = r"does not exist"
ERROR_NO_SUCH_ENTITY = r"NoSuchEntity"


def setup_connection(connection_settings):
    profile_name = util.get_connection_profile(connection_settings, "aws")
    os.environ["AWS_PROFILE"] = profile_name

    # check named profile exists
    try:
        _ = boto3.Session(profile_name=profile_name)
    except botocore.exceptions.ProfileNotFound:
        raise RuntimeError(f"[AWS] No such profile: {profile_name} (aws configure --profile {profile_name})")


def get_boto3_session():
    """get a boto3 session configured with the requested profile or bomb out"""
    # profile_name = util.get_connection_profile(connection, "aws")
    # try:
    #     session = boto3.Session(profile_name=profile_name)
    # except botocore.exceptions.ProfileNotFound:
    #     raise RuntimeError(f"[AWS] No such profile: {profile_name} (aws configure --profile {profile_name})")
    # return session

    return boto3.Session()


def get_eksctl_cmd():
    # profile_name = util.get_connection_profile(connection, "aws")
    # return ["eksctl", "--profile", profile_name]
    return ["eksctl"]


def route_table_for_subnet(ec2, subnet_id):
    # get the route tables
    logger.debug(f"looking describe_route_table subnet_id:{subnet_id}")
    response = ec2.describe_route_tables(
        Filters=[{
            'Name': 'association.subnet-id',
            'Values': [subnet_id]
        }]
    )
    try:
        logger.debug(f"...result: {response}")
        route_table_id = response["RouteTables"][0]["RouteTableId"]
    except KeyError as e:
        logger.warn(f"aws - no RouteTableId found for subnet {subnet_id} - associated?")
        route_table_id = None

    return route_table_id


def eks_cluster_info(cluster_name, data):
    sanity_check(data)

    # eksctl cluster info --> databag
    logger.debug("eks cluster info")
    eksctl_data = util.run_cmd_json(
        get_eksctl_cmd() + ["get", "cluster", cluster_name, "--output", "json"],
        data
    )

    flattened_eksctl_data = flatten_nested_dict(eksctl_data)
    logger.debug(f"loaded items:{len(flattened_eksctl_data)} data:{flattened_eksctl_data}")
    data.update(flattened_eksctl_data)

    #   + cluster_vpc_cidr
    #   + cluster_private_subnets
    #   + cluster_private_subnet_{n}
    #   + cluster_private_route_tables
    #   + cluster_private_route_table_{n}
    #   + cluster_public_subnets
    #   + cluster_public_subnet_{n}
    #   + cluster_public_route_tables
    #   + cluster_public_route_table_{n}
    ec2 = get_boto3_session().client('ec2')
    try:

        # VPC CIDR block
        response = ec2.describe_vpcs(
            Filters=[{
                'Name': 'vpc-id',
                'Values': [data['resourcesvpcconfig_vpcid']]
            }]
        )
        vpc_cidr_value = response["Vpcs"][0]["CidrBlock"]
        data["cluster_vpc_cidr"] = vpc_cidr_value

        # subnets
        response = ec2.describe_subnets(
            Filters=[{
                'Name': 'vpc-id',
                'Values': [data['resourcesvpcconfig_vpcid']]
            }]
        )

        private_subnets = list(filter(lambda x: x["MapPublicIpOnLaunch"], response["Subnets"]))
        private_subnet_ids = list(map(lambda x: x["SubnetId"], private_subnets))
        public_subnets = list(filter(lambda x: not x["MapPublicIpOnLaunch"], response["Subnets"]))
        public_subnet_ids = list(map(lambda x: x["SubnetId"], public_subnets))
        # join on whitespace - default separator for bash and python
        data["cluster_private_subnets"] = private_subnet_ids
        data["cluster_public_subnets"] = public_subnet_ids

        #   + cluster_public_route_tables
        #   + cluster_public_route_table_{n}
        data["cluster_public_route_tables"] = []
        for i, value in enumerate(public_subnet_ids):
            route_table_id = route_table_for_subnet(ec2, public_subnet_ids[i])
            data["cluster_public_route_tables"].append(route_table_id)
            data[f"cluster_public_route_table{i+1}"] = route_table_id

        #   + cluster_private_route_tables
        #   + cluster_private_route_table_{n}
        data["cluster_private_route_tables"] = []
        for i, value in enumerate(private_subnet_ids):
            route_table_id = route_table_for_subnet(ec2, private_subnet_ids[i])
            data["cluster_private_route_tables"].append(route_table_id)
            data[f"cluster_private_route_table{i+1}"] = route_table_id


        # flattened subnet ids for cloudformation/bash
        for i, value in enumerate(private_subnet_ids):
            data[f"cluster_private_subnet{i+1}"] = value

        for i, value in enumerate(public_subnet_ids):
            data[f"cluster_public_subnet{i+1}"] = value

        # if we are using ODIC to create cloudformation IAM roles, we need to
        # strip `https://` from the start of identity_oidc_issuer or we get
        # errors. Its impossible to do this natively in cloudformation so
        # simple munge here
        if "identity_oidc_issuer" in data:
            data["identity_oidc_issuer_stripped"] = \
                data["identity_oidc_issuer"].replace("https://", "")

    except KeyError as e:
        raise RuntimeError(
            f"missing required field in in databag: {e} - EKS cluster created yet?"
        )
    except IndexError as e:
        raise RuntimeError(
            f"No data returned for this EKS cluster: {e} - EKS cluster created yet?"
        )


def filename_to_stack_name(cloudformation_file):
    return os.path.basename(cloudformation_file)\
        .replace(constants.PATTERN_LOCAL_CLOUDFORMATION_FILE, "")\
        .replace(".yaml", "")


def cloudformation_param(key, value):
    return {
        "ParameterKey": key,
        "ParameterValue": value
    }


def stack_params(filename, data):
    # we can only insert stack parameters that are defined in YAML or cfn
    # will barf, so grab the parameter names from the CFN file and then
    # grab the corresponding parameters from the databag. If anything is
    # missing bomb out now before CFN does.
    # parse with cfn_flip as pyyaml cant handle things like `!Ref`
    # https://stackoverflow.com/a/55349491/3441106
    logger.debug(f"reading cloudformation parameters from {filename}")
    parsed = load_yaml(pathlib.Path(filename).read_text())

    params = []

    if "Parameters" in parsed:
        for cfn_param in parsed["Parameters"]:
            logger.debug(f"processing parameter: {cfn_param}")

            snakecase_param = util.string_to_snakecase(cfn_param)
            if data.get(cfn_param):
                value = data.get(cfn_param)
                logger.debug(f"{cfn_param} set to {value}")
                params.append(cloudformation_param(cfn_param, value))
            elif data.get(snakecase_param):
                value = data.get(snakecase_param)
                logger.debug(f"{cfn_param} set to {value}")
                params.append(cloudformation_param(cfn_param, value))
            elif "Default" in parsed["Parameters"][cfn_param]:
                # cloudformation allows the empty string as a default
                logger.debug(f"Using cloudformation default for {cfn_param}")
            else:
                raise KeyError(
                    f"Cloudformation file:{filename} missing param:{cfn_param} "
                        f"expected databag:{snakecase_param}, avaiable:{data.keys()}"
                )
    else:
        logger.debug(f"no Parameters section in {filename}")

    return params


def stack_exists(client, stack_name):
    logger.debug(f"cloudformation - describe_stacks: {stack_name}")
    try:
        response = client.describe_stacks(
            StackName=stack_name,
        )
        logger.debug(f"cloudformation - nothing thrown, response: {response}")
        exists = True
    except botocore.exceptions.ClientError as e:
        logger.debug(f"cloudformation exception - missing stack: {e}")
        if re.search(ERROR_MISSING, str(e), re.IGNORECASE):
            # normal AWS response for a missing stack
            exists = False
        else:
            # no idea... AWS broken?
            raise e

    logger.debug(f"cloudformation stack:{stack_name} exists:{exists}")
    return exists


def cloudformation_outputs(client, stack_name, data):
    prefixed_stack_name = get_prefixed_stack_name(stack_name, data)
    response = client.describe_stacks(
        StackName=prefixed_stack_name,
    )
    intermediate_databag = {}
    if "Stacks" in response and "Outputs" in response["Stacks"][0]:
        outputs = response["Stacks"][0]["Outputs"]
        logger.debug(f"cloudformation - checking outputs: {outputs}")
        for output in outputs:
            # replace the value of `{prefixed_stack_name}_` with `{stack_name}_`
            # eg foo-infra-efs --> infa_efs
            # this lets us reference variables in the stack without having to
            # adjust the name each time and avoids tricks like eval in bash.
            # the downside of this is you need to know what stack generated the
            # output name
            key_name = output["ExportName"].replace(f"{prefixed_stack_name}-", f"{stack_name}-")
            key_name = util.string_to_snakecase(key_name)

            string_value = str(output["OutputValue"])
            intermediate_databag[key_name] = string_value

    logger.debug(f"cloudformation - outputs:{len(intermediate_databag)} value: {intermediate_databag}")
    data.update(intermediate_databag)


# Cloudformation has a hard 51200 byte max file size. To work around this
# files must be hosted on S3. This is an issue with the AWS best
# practice/quickstart files...
def do_remote_cloudformation(working_dir, filename, verb, data):
    # local file contains a link to the S3 hosted cloudformation
    config = util.read_yaml_file(filename)

    remote = config["remote"]
    local_file = os.path.join(
        constants.PROCESSED_DIR,
        os.path.dirname(filename),
        config["local_file"]
    )

    # download the remote file so that:
    #   a) we can look at it
    #   b) we have a record of what was deployed
    #   c) we can grab the parameter list
    # download a fresh copy every time since cloudformation will...
    logger.debug(f"download {remote} to {local_file}")
    util.download(remote, local_file)
    stack_name = filename_to_stack_name(local_file)
    cloudformation(stack_name, local_file, verb, data, template_url=remote)


def do_local_cloudformation(working_dir, filename, verb, data):
    logger.info(f"cloudformation {filename}")
    template_body = pathlib.Path(filename).read_text()
    stack_name = filename_to_stack_name(filename)

    try:
        cloudformation(stack_name, filename, verb, data, template_body=template_body)
    except KeyError as e:
        if verb == constants.DOWN_VERB:
            logger.warning(f"missing values prevent running cloudformation - skipping as system going down: {e}")
        else:
            raise e


def get_prefixed_stack_name(stack_name, data):
    return f"{data['name']}-{stack_name}"


def cloudformation(stack_name, filename, verb, data, template_body=None, template_url=None):
    sanity_check(data)

    prefixed_stack_name = get_prefixed_stack_name(stack_name, data)
    params = stack_params(filename, data)
    client = get_boto3_session().client('cloudformation')
    exists = stack_exists(client, prefixed_stack_name)

    if template_body:
        template_source = {
            "TemplateBody": template_body,
        }
    elif template_url:
        template_source = {
            "TemplateURL": template_url
        }
    else:
        raise RuntimeError(
            "cloudformation - missing both TemplateBody and TemplateURL"
        )

    act = False

    if exists and verb == constants.UP_VERB:
        # update
        def ensure_fn():
            return client.update_stack(
                StackName=prefixed_stack_name,
                Parameters=params,
                Capabilities=['CAPABILITY_NAMED_IAM'],
                **template_source
            )

        waiter_name = "stack_update_complete"
        act = True
    elif exists and verb == constants.DOWN_VERB:
        # delete
        def ensure_fn():
            return client.delete_stack(
                StackName=prefixed_stack_name,
            )
        waiter_name = "stack_delete_complete"
        act = True
    elif not exists and verb == constants.UP_VERB:
        # create
        def ensure_fn():
            return client.create_stack(
                StackName=prefixed_stack_name,
                Parameters=params,
                Capabilities=['CAPABILITY_NAMED_IAM'],
                **template_source
            )
        waiter_name = "stack_create_complete"
        act = True
    elif not exists and verb == constants.DOWN_VERB:
        # already deleted
        logger.info(constants.MSG_UP_TO_DATE)
    else:
        raise RuntimeError(f"bad arguments in run_cloudformation - exists:{exists} verb:{verb}")

    if act:
        logger.debug(
            f"cloudformation stack:{prefixed_stack_name} file: {filename}"
        )

        # do the deed...
        try:
            with Halo(text=f"Cloudformation {stack_name}", spinner='dots'):
                response = ensure_fn()
                logger.debug(f"response: {response}")

                # ...wait for the result
                waiter = client.get_waiter(waiter_name)
                waiter.wait(
                    StackName=prefixed_stack_name,
                )

        # boto3 exceptions...
        # https://github.com/boto/botocore/blob/develop/botocore/exceptions.py
        except botocore.exceptions.ClientError as e:
            logger.debug(f"boto/client exception: {e}")
            if re.search(ERROR_UP_TO_DATE, str(e), flags=re.IGNORECASE):
                logger.info(constants.MSG_UP_TO_DATE)
            else:
                # no idea
                raise e
        except botocore.exceptions.WaiterError as e:
            logger.debug(f"botocore/waiter exception: {e}")
            if re.search(ERROR_AWS, str(e), flags=re.IGNORECASE):
                raise RuntimeError(f"cloudformation failed - check stack {prefixed_stack_name} in the AWS console")
            else:
                # no idea
                raise e

    # ...If we're still here our stack is up, grab all the cloudformation
    # outputs and add them to the databag
    if verb != constants.DOWN_VERB:
        cloudformation_outputs(client, stack_name, data)


def do_iam_policy(working_dir, filename, verb, data):
    logger.info(f"AWS IAM policy: {filename}")

    basename = os.path.basename(filename)
    policy_name = basename[:-len(constants.PATTERN_AWS_IAM_POLICY)]
    policy_arn = f"arn:aws:iam::{data['aws_account_id']}:policy/{policy_name}"
    client = get_boto3_session().client('iam')

    try:
        policy_exists = client.get_policy(PolicyArn=policy_arn)
    except botocore.exceptions.ClientError as e:
        logger.debug(f"AWS IAM - exception: {e}")
        if re.search(ERROR_NO_SUCH_ENTITY, str(e), re.IGNORECASE):
            policy_exists = False
        else:
            raise e

    if verb == constants.UP_VERB and policy_exists:
        logger.info(constants.MSG_UP_TO_DATE)
    elif verb == constants.UP_VERB and not policy_exists:
        logger.debug(f"creating IAM policy:{policy_name} file:{filename}")
        with open(filename, 'r') as f:
            file_content = f.read()
        response = client.create_policy(
            PolicyName=policy_name,
            PolicyDocument=file_content,
        )
        logger.debug(f"...result: {response}")
    elif verb == constants.DOWN_VERB and policy_exists:
        logger.debug(f"delete all versions of IAM policy:{policy_name}...")
        # delete all versions of the policy and then the policy itself
        response = client.list_policy_versions(PolicyArn=policy_arn)

        for version_info in response["Versions"]:
            if not version_info["IsDefaultVersion"]:
                version_id = version_info["VersionId"]
                logger.debug(f"deleting IAM policy:{policy_arn} version:{version_id}...")
                response = client.delete_policy_version(
                    PolicyArn=policy_arn,
                    VersionId=version_id,
                )
                logger.debug(f"...result: {response}")

        logger.debug(f"deleting overall IAM policy:{policy_name}...")
        try:
            response = client.delete_policy(
                PolicyArn=policy_arn
            )
            logger.debug(f"...result: {response}")
        except botocore.exceptions.ClientError as e:
            if verb == constants.DOWN_VERB:
                logger.warning(f"Error deleting policy:{policy_arn} - continuing as system is going down")
            else:
                raise e
    elif verb == constants.DOWN_VERB and not policy_exists:
        logger.info(constants.MSG_UP_TO_DATE)
    else:
        raise RuntimeError(f"AWS IAM - invalid verb {verb}")

    if verb == constants.UP_VERB:
        databag_key = ("aws_iam_policy_" + snakecase.convert(policy_name))
        extra_data = {databag_key: policy_arn}
        logger.debug(f"added to databag:{databag_key}")
        data.update(extra_data)


def do_iam_role(working_dir, filename, verb, data):
    logger.info(f"AWS IAM role: {filename}")

    basename = os.path.basename(filename)
    name = basename[:-len(constants.PATTERN_AWS_IAM_ROLE)]
    client = get_boto3_session().client('iam')

    try:
        exists = client.get_role(RoleName=name)
    except botocore.exceptions.ClientError as e:
        logger.debug(f"AWS IAM - exception: {e}")
        if re.search(ERROR_NO_SUCH_ENTITY, str(e), re.IGNORECASE):
            exists = False
        else:
            raise e

    if verb == constants.UP_VERB and exists:
        logger.info(constants.MSG_UP_TO_DATE)
    elif verb == constants.UP_VERB and not exists:
        logger.debug(f"creating IAM role:{name} file:{filename}")
        with open(filename, 'r') as f:
            file_content = f.read()
        response = client.create_role(
            RoleName=name,
            AssumeRolePolicyDocument=file_content,
        )
        logger.debug(f"...result: {response}")
    elif verb == constants.DOWN_VERB and exists:
        logger.debug(f"deleting IAM policy:{name}")
        response = client.delete_role(
            RoleName=name
        )
        logger.debug(f"...result: {response}")
    elif verb == constants.DOWN_VERB and not exists:
        logger.info(constants.MSG_UP_TO_DATE)
    else:
        raise RuntimeError(f"AWS IAM - invalid verb {verb}")

    if verb == constants.UP_VERB:
        databag_key = ("aws_iam_role_" + snakecase.convert(name))
        extra_data = {databag_key: name}
        logger.debug(f"added to databag:{databag_key}")
        data.update(extra_data)


def sanity_check(data):
    if not data.get("aws_region"):
        raise RuntimeError("must set aws_region in databag")
    if not data.get("aws_account_id"):
        raise RuntimeError("must set aws_account_id in databag")


def secret_exists(client, secret_id):
    try:
        response = client.describe_secret(
            SecretId=secret_id
        )
        exists = True
        deleted = True if response.get("DeletedDate") else False
    except botocore.exceptions.ClientError as e:
        logger.error(f"--> {e}")
        exists = False
        deleted = False
    logger.debug(f"secret_id:{secret_id} exists:{exists}")
    return exists, deleted


def update_secret(client, secret_id, secret_value, deleted):
    if deleted:
        logger.debug(f"Restoring deleted secret: {secret_id}")
        _ = client.restore_secret(
            SecretId=secret_id
        )

    response = client.update_secret(
        SecretId=secret_id,
        SecretString=secret_value
    )
    logger.info(f"created secret id:{secret_id} ARN:{response}")


def create_secret(client, secret_id, secret_value):
    response = client.create_secret(
        Name=secret_id,
        SecretString=secret_value
    )
    logger.info(f"created secret id:{secret_id} ARN:{response}")


def delete_secret(client, secret_id):
    exists, deleted = secret_exists(client, secret_id)
    if exists and not deleted:
        response = client.delete_secret(
            SecretId=secret_id
        )
        logger.debug(f"secret id:{secret_id} deleted:{response}")
    else:
        logger.debug(f"secret id:{secret_id} does not exist, moving on")


def ensure_secret(data, verb, secret):
    client = get_boto3_session().client('secretsmanager')

    exists, deleted = secret_exists(client, secret["name"])
    if verb == constants.UP_VERB and exists:
        update_secret(client, secret["name"], secret["value"], deleted)
    elif verb == constants.UP_VERB and not exists:
        create_secret(client, secret["name"], secret["value"])
    elif verb == constants.DOWN_VERB and exists:
        delete_secret(client, secret["name"])
    elif verb == constants.DOWN_VERB and not exists:
        logger.debug(f"secretsmanager - already deleted:{secret['name']}")
    else:
        raise RuntimeError(f"secretsmanager - invalid verb: {verb}")


# `secret_id` - The identifier of the secret whose details you want to
# retrieve. You can specify either the Amazon Resource Name (ARN) or the
# friendly name of the secret.
def do_secrets_manager(working_dir, filename, verb, data):
    logger.info(f"secretsmanager: {filename}")
    processed_file = util.substitute_placeholders_from_file_to_file(working_dir, filename, "#", verb, data)
    config = util.read_yaml_file(processed_file)

    for secret in config.get("secrets", []):
        ensure_secret(data, verb, secret)

    # munged secrets files must not be left on disk
    logger.debug(f"delete file that may contain secrets: {processed_file}")
    os.unlink(processed_file)


def do_eksctl(working_dir, filename, verb, data):
    logger.info(f"eksctl: ${filename}")
    processed_filename = util.substitute_placeholders_from_file_to_file(working_dir, filename, "#", verb, data)
    config = util.read_yaml_file(processed_filename)

    # eksctl needs cluster name - grab this well-known fields from
    # processed yaml file to avoid needing well-known keys in databag
    try:
        cluster_name = config["metadata"]["name"]
    except KeyError as e:
        raise RuntimeError(f"eksctl file:{filename} missing required value {e}")

    try:
        util.run_cmd(get_eksctl_cmd() + ["get", "cluster", "-n", cluster_name], data)
        exists = True
    except RuntimeError as e:
        logger.debug(f"cluster probably doesnt exist: {e}")
        exists = False

    if verb == constants.UP_VERB and exists:
        logger.info(constants.MSG_UP_TO_DATE)
    elif verb == constants.UP_VERB and not exists:
        util.run_cmd(get_eksctl_cmd() + ["create", "cluster", "-f", processed_filename], data)
    elif verb == constants.DOWN_VERB and exists:
        util.run_cmd(get_eksctl_cmd() + ["delete", "cluster", "--name", cluster_name], data)
    elif verb == constants.DOWN_VERB and not exists:
        logger.info(constants.MSG_UP_TO_DATE)
    else:
        raise RuntimeError(f"eksctl - invalid verb: {verb}")

    if verb == constants.UP_VERB:
        eks_cluster_info(cluster_name, data)


def eks_name_to_kubectl_context_id(iam_user, cluster_name, region):
    return f"{iam_user}@{cluster_name}.{region}.eksctl.io"


def boto3_version():
    return boto3.__version__



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ringmaster']

package_data = \
{'': ['*'], 'ringmaster': ['res/*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'boto3>=1.17.24,<2.0.0',
 'cfn-flip>=1.2.3,<2.0.0',
 'docopt>=0.6.2,<0.7.0',
 'halo>=0.0.31,<0.0.32',
 'loguru>=0.5.3,<0.6.0',
 'python-cloudflare>=1.0.1,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'snakecase>=1.0.1,<2.0.0',
 'snowflake-connector-python>=2.4.1,<3.0.0']

entry_points = \
{'console_scripts': ['ringmaster = ringmaster.cli:main']}

setup_kwargs = {
    'name': 'ringmaster.show',
    'version': '0.4.1',
    'description': 'The world is a circus and you are the ringmaster!',
    'long_description': "# Ringmaster\n```\n         _\n       _[_]_\n       _(_)______.-'`-.\n      /, >< ,----'     `-._.-'*\n      \\\\|::|  Welcome to the Circus\n        |/\\|  We already got enough Clowns,\n        ||||  You got any experiance with\n        ||||  Being shot from a canon??\n     __(_/\\_)\n    /`-..__.,-'\\\n   /   __/\\__   \\\n   `._ \\    / _.'MJP\n      ``|/\\|-'\n```\n\nRingmaster organises a bunch of other tools on your behalf so that you don't\nhave to. The aim is you can create, updated and delete entire stacks crossing\ncloudformation, EKS, kubectl, helm and random Python/BASH scripts with a single\ncommand.\n\nRingmaster helps you create and share your automation scripts with others, so\nyou can get up and running as quick as possible. There are no agents, hubs, \ngits or daemons - unless you add them yourself. \n\nThere is also no custom DSL or new programming language to learn, although\n[jinja2](https://jinja.palletsprojects.com/) is used for templating.\n\nRingmaster is just files on a disk and calls to other systems made in an order\nyou decide.\n\n## How does it work?\n\nYou create a directory of scripts to process, like this:\n\n```\nstack/\n├── 0010-iam\n│     ├── AWSLoadBalancerController.iam_policy.json\n│     ├── Certbot.iam_policy.json\n│     ├── EksDeploy.iam_policy.json\n│     ├── EksExternalSecrets.iam_policy.json\n│     ├── ExternalDns.iam_policy.json\n│     └── metadata.yaml\n├── 0020-efs\n│     ├── efs.cloudformation.yaml\n│     └── metadata.yaml\n├── 0030-vpc\n│     ├── metadata.yaml\n│     ├── vpc.remote_cloudformation.yaml\n│     └── vpc.yaml\n...\n```\n\nThen you run `ringmaster` like this:\n\n`ringmaster stack up`\n\nRingmaster will carry out the _create_ action of each script, running each \nscript in alphabetical order by _directory_ and then _file_\n\n`ringmaster stack down`\n\nRingmaster will carry out the delete action of each script, in _reverse_\nalphabetical order\n\n**The `up` and `down` actions are \n[idempotent](https://en.wikipedia.org/wiki/Idempotence#Computer_science_examples)\nso you can run them as many times as you like**\n\n\n## What's in the scripts? do I have to learn a new language?\n\nNo! The scripts use the languages and tools you already know and love, eg:\n\n* Cloudformation\n* Bash\n* Python\n* Kubernetes deployment descriptors\n* ...etc\n\nRingmaster uses a [databag](doc/concepts.md#databag) to give each script the\nright inputs and collects outputs that may be required later. Combined with a\nsimple built-in variable substitution system, this makes gluing completely \ndifferent systems together easy, eg:\n\n```\ncloudformation -> ekscl -> more cloudformation -> heml -> kubectl -> ...\n```\n\nTo reduce dependency on ringmaster and allow easy debugging and repeatable\ndeployments, substitution results are stored adjacent to their input files, so\nthey can be added to git or use directly by tools such as `kubectl`.\n\n## Reference\n\n1. [Concepts](doc/concepts.md)\n2. [Setup](doc/setup.md)\n3. [Authentication](doc/authentication.md)\n4. [Handlers](doc/handlers.md)\n5. [Scripts](doc/scripts.md)\n6. [Variables](doc/variables.md)   \n7. [Worked Example](doc/worked_example.md)\n",
    'author': 'Geoff Williams',
    'author_email': 'geoff@declarativesystems.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://ringmaster.show',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tempfile
import os
import yaml
import shutil
from loguru import logger
from .util import run_cmd
from pathlib import Path
from ringmaster import constants
import ringmaster.util as util
import re

kubectl_context = None


def setup_connection(connection_settings):
    global kubectl_context
    kubectl_context = util.get_connection_profile(connection_settings, "k8s")
    logger.debug(f"k8s context set to: {kubectl_context}")


def get_kubectl_cmd():
    """get the kubectl command"""
    return ["kubectl", "--context", kubectl_context]


def get_helm_cmd():
    """get the kubectl command to run with context set or bomb out"""
    return ["helm", "--kube-context", kubectl_context]


def check_kubectl_session():
    """check kubectl connected to cluster before running commands"""
    try:
        run_cmd(get_kubectl_cmd() + ["get", "svc"])
        connected = True
    except RuntimeError:
        connected = False
    return connected


def run_kubectl(verb, flag, path, data):
    if data is None:
        data = {}

    if verb == constants.UP_VERB:
        action = "apply"
    elif verb == constants.DOWN_VERB:
        action = "delete"
    else:
        raise ValueError(f"invalid verb: {verb}")

    if check_kubectl_session():
        # --force is to recreate any immutable resources we touched
        cmd = get_kubectl_cmd() + [action, "--force", flag, path]
        if data.get("debug"):
            cmd.append("-v=2")

        try:
            run_cmd(cmd, data)
        except RuntimeError as e:
            if verb == constants.DOWN_VERB:
                logger.warning("Error running kubectl but system is going down - ignoring")
            else:
                raise e
    elif verb == constants.DOWN_VERB:
        logger.warning("kubectl has no current context but system is going down - ignoring")
    else:
        raise RuntimeError("kubectl not logged in - context is not set")


def do_kubectl(working_dir, filename, verb, data=None):
    # substitute ${...} variables from databag, bomb out if any missing
    logger.info(f"kubectl: {filename}")

    try:
        processed_file = util.substitute_placeholders_from_file_to_file(
            working_dir,
            filename,
            "#",
            verb,
            data
        )
        logger.debug(f"kubectl processed file: {processed_file}")

        run_kubectl(verb, "-f", processed_file, data)
    except RuntimeError as e:
        if verb == constants.DOWN_VERB:
            logger.warning(f"kubectl error - moving on: {e}")
        else:
            raise e
    except KeyError as e:
        if verb == constants.DOWN_VERB:
            logger.warning(f"missing key - moving on: {e}")
        else:
            raise e


def do_kustomizer(working_dir, filename, verb, data=None):
    logger.info(f"kustomizer: {filename}")
    sources_dir = os.path.dirname(filename)
    try:
        run_kubectl(verb, "-k", sources_dir, data)
    except RuntimeError as e:
        if verb == constants.DOWN_VERB:
            logger.error(f"kustomizer error - moving on: {e}")
        else:
            raise e


def helm_repos(base_cmd, config, filename):
    repos_list = util.run_cmd_json(base_cmd + ["repo", "list", "--output", "json"])
    logger.debug(f"helm repos installed: {repos_list}")
    repos_installed = list(map(lambda x: x["name"], repos_list))
    logger.debug(f"helm repos installed: {repos_installed}")

    repos = config.get("repos", {})
    if repos:
        for repo_key, source in config.get("repos", {}).items():
            logger.debug(f"helm - request install repo:{repo_key}")
            if repo_key not in repos_installed:
                logger.info(f"helm - installing repo:{repo_key}")
                cmd = base_cmd + ["repo", "add", repo_key, source]
                util.run_cmd(cmd)
    else:
        logger.warning(f"helm - no helm repos specified in {filename}")


def do_helm(working_dir, filename, verb, data=None):
    logger.info(f"helm: {filename}")

    # settings for helm...
    processed_filename = util.substitute_placeholders_from_file_to_file(
        working_dir,
        filename,
        "#",
        verb,
        data
    )
    config = util.read_yaml_file(processed_filename)

    # if there is an adjacent `values.yaml` file, process it for substitutions and use it
    values_yaml = os.path.join(os.path.dirname(filename), "values.yaml")
    if os.path.exists(values_yaml):
        processed_values_file = util.substitute_placeholders_from_file_to_file(
            working_dir,
            values_yaml,
            "#",
            verb,
            data
        )
        logger.debug(f"helm - values: {processed_values_file}")
    else:
        processed_values_file = False

    if verb == constants.UP_VERB:
        helm_command = "install"
    elif verb == constants.DOWN_VERB:
        helm_command = "uninstall"
    else:
        raise RuntimeError(f"helm - invalid verb: {verb}")

    base_cmd = get_helm_cmd()
    if data.get("debug"):
        base_cmd.append("--debug")

    if config.get("namespace"):
        namespace = config['namespace']
        logger.debug(f"using namespace {namespace}")
        base_cmd.append("-n")
        base_cmd.append(namespace)

    try:
        # helm repos...
        if verb == constants.UP_VERB:
            helm_repos(base_cmd, config, filename)

        # helm deployments
        helm_list = util.run_cmd_json(base_cmd + ["list", "--output", "json"])
        helm_deployments = list(map(lambda x: x["name"], helm_list))
        logger.debug(f"helm deployments installed: {helm_deployments}")
        exists = config["name"] in helm_deployments
        if (verb == constants.UP_VERB and exists) or (verb == constants.DOWN_VERB and not exists):
            logger.info(f"helm - already installed:{config['name']}")
        elif (verb == constants.UP_VERB and not exists) or (verb == constants.DOWN_VERB and exists):
            logger.info(f"helm - {helm_command}:{config['name']}")
            cmd = base_cmd + [helm_command, config["name"]]
            if verb == constants.UP_VERB:
                cmd.append(config["install"])
                if processed_values_file:
                    cmd.append("--values")
                    cmd.append(processed_values_file)
                else:
                    for setting in config.get("set", []):
                        cmd.append("--set")
                        cmd.append(setting)
                cmd += config.get("options", [])
                version = config.get("version")
                if version:
                    cmd += ["--version", version]
                else:
                    logger.warning(f"recommending versioning helm chart in {filename}")
                cmd += config.get("options", [])
            util.run_cmd(["helm", "repo", "update"])
            util.run_cmd(cmd)
        else:
            raise RuntimeError(f"helm - invalid verb {verb}")

    except KeyError as e:
        raise RuntimeError(f"helm - {filename} missing key: {e}")
    except RuntimeError as e:
        if verb == constants.DOWN_VERB:
            logger.error(f"helm - error running helm, moving on: {e}")
        else:
            raise e


def do_secret_kubectl(working_dir, filename, verb, data):
    """create or delete a secret from a kubectl template file. This results in
    calling the k8s api directly - no processed file (which would contain the
    secret...) is created

    The input yaml file must be formatted like this:

    ---
    apiVersion: v1
    kind: Secret
    metadata:
      name: ...
    ---
    data:
        keyname: value (we will base64 encode it for you)
    """
    logger.info(f"secret_kubectl: {filename}")

    # step 1 - read entire file into memory and split on yaml record separator
    # `---` make sure we have exactly 2 records
    with open(filename) as f:
        file_contents = f.read()

    records = re.split(r'\s*---\s*', file_contents)


    # step 2 - process all substitutions in first document (metadata)
    # since yaml lets the file start with `---` its possible to end up
    # with a 'blank' first record so skip until we find something
    yaml_data = False
    while not yaml_data:
        yaml_string_metadata = util.substitute_placeholders_from_memory_to_memory(records[0], verb, data)

        # step 3 - convert string to yaml data structure - this will be merged
        # with the looked-up data from record 2 to build the entire secret
        yaml_data = yaml.safe_load(yaml_string_metadata)

        if not yaml_data:
            # skip this blank record and try again...
            records.pop(0)

    record_count = len(records)
    if record_count == 2:
        # secret file in correct format

        # step 4 - record 2 contains the secret data to add - if we are upping
        # substitute in the real values
        if verb == constants.UP_VERB:
            # substitute function expects list of strings...
            yaml_string_secret = util.substitute_placeholders_from_memory_to_memory(records[1], verb, data)
            logger.debug(f"raw secret data after placeholder substitution: {yaml_string_secret}")
            yaml_data_secret = yaml.safe_load(yaml_string_secret)

            # parsed yaml must contain `data` key...
            secret_data = yaml_data_secret.get("data")
            if secret_data:
                # create the `data` key in the overall structure if needed
                if not yaml_data.get("data"):
                    yaml_data["data"] = {}

                # base 64 encode each field and add it to the yaml secret
                for k, v in secret_data.items():
                    logger.debug(f"secret_kubectl - adding data field:{k} ")
                    yaml_data["data"][k] = util.base64encode(v)
            else:
                raise RuntimeError(f"secret_kubectl - record 2 missing key `data` filename:${filename}")

        else:
            # Add dummy section to keep kubectl from complaining
            yaml_data["data"] = {}

        # step 5 - secret completed - save it somewhere, kubectl, delete
        _, secret_file = tempfile.mkstemp(suffix=".yaml", prefix="ringmaster")
        util.save_yaml_file(secret_file, yaml_data)

        logger.debug("secret_kubectl - creating secret with kubectl")

        # print the entire secret for debug purposes ;-)
        logger.debug(f"secret: {yaml_data}")

        run_kubectl(verb, "-f", secret_file, data)
        os.unlink(secret_file)
    else:
        raise RuntimeError(f"secret_kubectl - ${filename} must contain exactly 2 documents, found:{record_count}")


# Copyright 2020 Declarative Systems Pty Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ringmaster

Usage:
  ringmaster [--debug] <dir> (up|down) [--start=<dir>] [--env=<dir>] [--no-merge-env]
  ringmaster [--debug] get <dir> <url>
  ringmaster [--debug] metadata <dir> [--include=<files>]
  ringmaster [--debug] --run <filename> (up|down) [--env=<dir>] [--no-merge-env]
  ringmaster --version

Options:
  -h --help         Show this screen.
  --version         Show version.
  --debug           Extra debugging messages
  --env=<dir>       Read/write databags to this sub-directory under .env
                    otherwise just use .env. Databags will be merged with any
                    databags in the parent directory with child values taking
                    precedence unless --no-merge-env is used
  --no-merge-env    Do not merge databag values between env directories
  --start=<dir_num> up: start here count up, down: start here count down
  --include=<files> comma delimited list of extra files to add to metadata
"""

from loguru import logger
import sys
from docopt import docopt
import ringmaster.api as api
import ringmaster.version as version
import ringmaster.constants as constants
import os

debug = False

# color logs
# https://stackoverflow.com/a/56944256/3441106
def setup_logging(level, logger_name=None):
    logger_name = logger_name or __name__.split(".")[0]
    log_formats = {
        "DEBUG": "{time}<level> {level} [{module}] {message}</level>",
        "INFO": "<level>[{module}] {message}</level>",
    }


    # custom level for program output so it can be nicely colourised
    logger.remove()
    logger.debug(f"{logger_name} {level}")
    logger.add(sys.stdout, colorize=True, format=log_formats[level], level=level)
    prog_level = logger.level("OUTPUT", no=25, color="<white><dim>", icon="🤡")

    logger.debug("====[debug mode enabled]====")


def main():
    arguments = docopt(__doc__, version=version.__version__)
    setup_logging("DEBUG" if arguments['--debug'] else "INFO")
    api.debug = arguments['--debug']
    logger.debug(f"parsed arguments: ${arguments}")
    merge = not arguments.get("--no-merge-env")
    env_name = arguments["--env"]

    try:
        if arguments["down"]:
            verb = constants.DOWN_VERB
        elif arguments["up"]:
            verb = constants.UP_VERB
        elif arguments["get"]:
            verb = constants.GET_VERB
        elif arguments["metadata"]:
            verb = constants.METADATA_VERB
        else:
            raise RuntimeError("one of (up|down|get) is required")

        api.system_info()
        working_dir = os.getcwd()
        if arguments["get"]:
            api.get(arguments["<dir>"], arguments["<url>"])
        elif arguments["metadata"]:
            api.write_metadata(arguments["<dir>"], arguments.get("--include", []))
        elif arguments["<dir>"]:
            api.run_dir(working_dir, arguments["<dir>"], merge, env_name, arguments['--start'], verb)
        elif arguments["--run"]:
            api.run(working_dir, arguments['<filename>'], merge, env_name, verb)
        else:
            raise RuntimeError("one of <dir> or --run expected")

    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        logger.error(str(exc_value))
        if arguments['--debug']:
            logger.exception(e)
        sys.exit(1)

